/**
 * Package that contains interfaces needed for dynamic, pluggable
 * format (auto)detection; as well as basic utility classes for
 * simple format detection functionality.
 */
package com.fasterxml.jackson.core.format;
